# -*- coding: utf-8 -*-
"""
rio.app
~~~~~~~~~

Implement of rio app factory based on Flask.
"""

from flask import Flask

from rio.setup import configure_app
from rio.setup import init_core
from rio.setup import register_blueprints

def create_app():
    """Flask application factory function."""
    app = Flask(__name__)
    app.config_from_envvar = app.config.from_envvar
    app.config_from_object = app.config.from_object
    configure_app(app)
    init_core(app)
    register_blueprints(app)
    return app



# -*- coding: utf-8 -*-
"""
rio.utils.http
~~~~~~~~~~~~~~

"""
import warnings
from urllib import urlencode

import six
import requests
from requests.exceptions import SSLError
from flask import current_app

from rio.core import celery
from rio.signals import webhook_ran
# In case SSL is unavailable (light builds) we can't import this here.
try:
    from OpenSSL.SSL import ZeroReturnError
except ImportError:
    class ZeroReturnError(Exception):
        pass


def get_user_agent():
    return 'rio/%s' % current_app.config.get('RIO_VERSION')

def build_session():
    """
    Rio requests session will attach rio identity in header.
    """
    session = requests.Session()
    session.headers.update({'User-Agent': get_user_agent()})
    return session


def raven_context(url, method=None, params=None, data=None, json=None,
                  headers=None, allow_redirects=False, timeout=30,
                  verify_ssl=True, user_agent=None):
    headers = {
        'User-Agent': get_user_agent(),
    }

    if json:
        headers.setdefault('Content-Type', 'application/json')

    if params:
        query_string = urlencode(params)
    else:
        query_string = None

    if json:
        data = json

    if not method:
        method = 'POST' if (data or json) else 'GET'

    return {
        'method': method,
        'url': url,
        'query_string': query_string,
        'data': data,
        'headers': headers,
    }


def urlopen(url, method=None, params=None, data=None, json=None,
            headers=None, allow_redirects=False, timeout=30,
            verify_ssl=True, user_agent=None):
    """
    A slightly safer version of ``urlib2.urlopen`` which prevents redirection
    and ensures the URL isn't attempting to hit a blacklisted IP range.
    """
    if user_agent is not None:
        warnings.warn('user_agent is no longer used with safe_urlopen')

    session = build_session()

    kwargs = {}

    if json:
        kwargs['json'] = json
        if not headers:
            headers = {}
        headers.setdefault('Content-Type', 'application/json')

    if data:
        kwargs['data'] = data

    if params:
        kwargs['params'] = params

    if headers:
        kwargs['headers'] = headers

    if method is None:
        method = 'POST' if (data or json) else 'GET'

    try:
        response = session.request(
            method=method,
            url=url,
            allow_redirects=allow_redirects,
            timeout=timeout,
            verify=verify_ssl,
            **kwargs
        )
    # Our version of requests does not transform ZeroReturnError into an
    # appropriately generically catchable exception
    except ZeroReturnError as exc:
        import sys
        exc_tb = sys.exc_info()[2]
        six.reraise(SSLError, exc, exc_tb)
        del exc_tb

    # requests' attempts to use chardet internally when no encoding is found
    # and we want to avoid that slow behavior
    if not response.encoding:
        response.encoding = 'utf-8'

    return response


def urlread(response):
    return response.content


def is_success_response(response):
    return 200 <= response.status_code < 300

def is_failure_response(response):
    return 500 <= response.status_code < 600

def is_invalid_response(response):
    return 400 <= response.status_code < 500

class FailureWebhookError(Exception):
    pass

class InvalidResponseError(FailureWebhookError):
    """The remote server gave an invalid response."""


class RemoteExecuteError(FailureWebhookError):
    """The remote task gave a custom error."""


class UnknownStatusError(FailureWebhookError):
    """The remote server gave an unknown status."""


def extract_response(raw_response):
    """Extract requests response object.

    only extract those status_code in [200, 300).

    :param raw_response: a requests.Resposne object.
    :return: content of response.
    """
    data = urlread(raw_response)

    if is_success_response(raw_response):
        return data
    elif is_failure_response(raw_response):
        raise RemoteExecuteError(data)
    elif is_invalid_response(raw_response):
        raise InvalidResponseError(data)
    else:
        raise UnknownStatusError(data)

def dispatch_webhook_request(url=None, method='GET', params=None,
                             json=None, headers=None, timeout=5):
    """Task dispatching to an URL.

    :param url: The URL location of the HTTP callback task.
    :param method: Method to use when dispatching the callback. Usually
        `GET` or `POST`.
    :param params: Keyword arguments to pass on to the HTTP callback.
    :param json: JSON as body to pass on to the POST HTTP callback.
    :param headers: HTTP headers applied to callback.
    """
    if method == 'GET':
        resp = urlopen(url, method, params=params, headers=headers)
    elif method == 'POST':
        resp = urlopen(url, method, json=json, headers=headers)
    else:
        raise NotImplementedError

    return extract_response(resp)



# -*- coding: utf-8 -*-

import logging
from uuid import uuid4
from os import environ

from pytest import fixture

from rio.app import create_app
from rio.core import db as db

@fixture(scope='session')
def app(request):
    environ['RIO_ENV'] = 'test'
    app_ = create_app()
    ctx = app_.app_context()
    ctx.push()
    request.addfinalizer(ctx.pop)
    return app_


@fixture(scope='session')
def database(request, app):
    uuid = uuid4().hex
    for table in db.metadata.sorted_tables:
        table.name = 'test_%s_%s' % (uuid, table.name)
    db.create_all()
    def fin():
        db.session.remove()
        db.drop_all()
        db.engine.dispose()
    request.addfinalizer(fin)
    return db

@fixture
def session(request, database):
    # cache.mc._flushall()
    def fin():
        for table in reversed(database.metadata.sorted_tables):
            database.session.execute(table.delete())
        database.session.commit()
        database.session.close()
        # cache.mc._flushall()
    request.addfinalizer(fin)
    return database.session

@fixture
def client(app, request):
    ctx = app.test_request_context()
    ctx.push()
    request.addfinalizer(ctx.pop)
    return app.test_client()

def pytest_addoption(parser):
    '''
    Add CLI options to py.test
    '''
    group = parser.getgroup('logging', 'Logging Configuration')
    group.addoption('--logging-level',
                    dest='logging_level',
                    default='INFO',
                    help='log format as used by the logging module')
    parser.addini('logging_level',
                  'log level as used by the logging module')

def pytest_configure(config):
    logging.root.addHandler(logging.StreamHandler())
    logging.root.setLevel(
        getattr(logging, config.getini('logging_level') or
                config.getvalue('logging_level')))



# -*- coding: utf-8 -*-
"""
rio.tasks
~~~~~~~~~~

Implement of rio tasks based on celery.
"""

from time import time

from celery import chord
from requests import ConnectionError

from celery.utils.log import get_task_logger

from rio.core import celery
from rio.core import sentry
from rio.utils.http import dispatch_webhook_request
from rio.utils.http import raven_context
from rio.utils.http import FailureWebhookError


logger = get_task_logger(__name__)


def _build_request_for_calling_webhook(event, webhook, payload):
    event_identity = 'uuid=%s,project=%s,action=%s' % (
        str(event['uuid']), event['project'], event['action']
    )

    request = {
        'url': webhook['url'],
        'method': webhook['method'],
        'headers': {
            'X-RIO-EVENT': event_identity,
        }
    }


    if webhook['method'] == 'GET':
        request['params'] = payload
    elif webhook['headers'].get('Content-Type') == 'application/json':
        request['json'] = payload
    else:
        request['data'] = payload

    return request


@celery.task()
def call_webhook(event, webhook, payload):
    """Build request from event,webhook,payoad and parse response."""
    started_at = time()

    logger.debug('REQUEST %(uuid)s %(method)s %(url)s %(payload)s' % dict(
        uuid=str(event['uuid']),
        url=webhook['url'],
        method=webhook['method'],
        payload=payload,
    ))

    request = _build_request_for_calling_webhook(event, webhook, payload)

    try:
        content = dispatch_webhook_request(**request)

        logger.debug('RESPONSE %(uuid)s %(method)s %(url)s %(data)s' % dict(
            uuid=str(event['uuid']),
            url=webhook['url'],
            method=webhook['method'],
            data=content,
        ))

        return dict(
            parent=str(event['uuid']),
            content=content,
            started_at=started_at,
            ended_at=time()
        )
    except (FailureWebhookError, ConnectionError) as exception:
        if sentry.client:
            http_context = raven_context(**request)
            sentry.captureException(data={'request': http_context})

        logger.error('RESPONSE %(uuid)s %(method)s %(url)s %(error)s' % dict(
            uuid=str(event['uuid']),
            method=webhook['method'],
            url=webhook['url'],
            error=exception.message,))

        return dict(
            parent=str(event['uuid']),
            error=exception.message,
            started_at=started_at,
            ended_at=time(),
        )


@celery.task()
def merge_webhooks_runset(runset):
    """Make some statistics on the run set.

    """
    min_started_at = min([w['started_at'] for w in runset])
    max_ended_at = max([w['ended_at'] for w in runset])
    ellapse = max_ended_at - min_started_at
    errors_count = sum(1 for w in runset if 'error' in w)
    total_count = len(runset)
    return dict(
        ellapse=ellapse,
        errors_count=errors_count,
        total_count=total_count,
    )


def exec_event(event, webhooks, payload):
    """Execute event.

    Merge webhooks run set to do some stats after all
    of the webhooks been responded successfully.

    +---------+
    |webhook-1+--------------------+
    +---------+                    |
                                   |
    +---------+                    |
    |webhook-2+-------------+      |
    +---------+             +------+-----+
                            |merge runset+------>
    +---------+             +------+-----+
    |webhook-3+-------------+      |
    +---------+                    |
                                   |
    +---------+                    |
    |...      +--------------------+
    +---------+


    Error webhook will be propagated. Note that other webhook
    calls will still execute.
    """
    calls = (
        call_webhook.s(event, webhook, payload)
        for webhook in webhooks
    )
    callback = merge_webhooks_runset.s()

    call_promise = chord(calls)
    promise = call_promise(callback)
    return promise



# -*- coding: utf-8 -*-

from datetime import datetime

from rio.core import db
from .utils import ins2dict

class Action(db.Model):
    """Action Model.

    Actions are emitted and subscribed as identity of event.
    """

    __tablename__ = 'action'
    __table_args__ = (
        db.UniqueConstraint('project_id', 'slug', name='ux_action_project_slug'),
    )

    id = db.Column(db.Integer(), primary_key=True)
    slug = db.Column(db.String(64), nullable=False)
    project_id = db.Column(db.Integer(), db.ForeignKey('project.id'), nullable=False)
    description = db.Column(db.String(255), nullable=False, default='')
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)

    webhooks = db.relationship('Webhook', backref=db.backref('action'), lazy='dynamic')

    def to_full_dict(self):
        data = ins2dict(self)
        project_id = data.pop('project_id')
        data['project'] = ins2dict(self.project, 'simple')
        data['webhooks'] = [ins2dict(webhook, 'simple') for webhook in self.webhooks]
        return data

    def to_simple_dict(self):
        return dict(
            slug=self.slug,
        )



# -*- coding: utf-8 -*-
"""
rio.models.utils
~~~~~~~~~~~~~~~~
"""

from uuid import UUID
from time import mktime
from datetime import datetime

from werkzeug.utils import import_string
from flask import abort


def _formatted(value):
    if isinstance(value, (int, str, long, )):
        return value
    if isinstance(value, datetime):
        return int(mktime(value.timetuple()))
    return value


def _turn_row_to_dict(row):
    return {
        column.name: _formatted(getattr(row, column.name))
        for column in row.__table__.columns
    }


def get_model(model):
    """Get a model class by model name.

    if model does not exist, this function will throw ImportError.

    :param model: a string of model name.
    :return: a SQLAlchemy Model.
    """
    return import_string('rio.models.%s.%s' % (model.lower(), model.capitalize()))


def ins2dict(ins, kind=''):
    """Turn a SQLAlchemy Model instance to dict.

    :param ins: a SQLAlchemy instance.
    :param kind: specify which kind of dict tranformer should be called.
    :return: dict, instance data.

    If model has defined `to_xxx_dict`, then ins2dict(ins, 'xxx') will
    call `model.to_xxx_dict()`. Default kind is ''.

    If model not defined 'to_dict', then ins2dict will transform according
    by model column definition.
    """
    if kind and hasattr(ins, 'to_%s_dict' % kind):
        return getattr(ins, 'to_%s_dict' % kind)()
    elif hasattr(ins, 'to_dict'):
        return getattr(ins, 'to_dict')()
    else:
        return _turn_row_to_dict(ins)


def get_instance(model, instance_id):
    """Get an instance by id.

    :param model: a string, model name in rio.models
    :param id: an integer, instance id.
    :return: None or a SQLAlchemy Model instance.
    """
    try:
        model = get_model(model)
    except ImportError:
        return None

    return model.query.get(instance_id)


def get_data_or_404(model, instance_id, kind=''):
    """Get instance data by id.

    :param model: a string, model name in rio.models
    :param id: an integer, instance id.
    :param kind: a string specified which kind of dict tranformer should be called.
    :return: data.
    """
    instance = get_instance(model, instance_id)

    if not instance:
        return abort(404)

    return ins2dict(instance, kind)


def get_instance_by_slug(model, slug, **kwargs):
    """Get an instance by slug.

    :param model: a string, model name in rio.models
    :param slug: a string used to query by `slug`. This requires there is a
                 slug field in model definition.
    :return: None or a SQLAlchemy Model instance.
    """
    try:
        model = get_model(model)
    except ImportError:
        return None

    query_params = dict(kwargs)
    query_params['slug'] = slug

    return model.query.filter_by(**query_params).first()


def get_data_by_slug_or_404(model, slug, kind='', **kwargs):
    """Get instance data by slug and kind. Raise 404 Not Found if there is no data.

    This function requires model has a `slug` column.

    :param model: a string, model name in rio.models
    :param slug: a string used to query by `slug`. This requires there is a
                 slug field in model definition.
    :param kind: a string specified which kind of dict tranformer should be called.
    :return: a dict.
    """

    instance = get_instance_by_slug(model, slug, **kwargs)

    if not instance:
        return abort(404)

    return ins2dict(instance, kind)


def get_instance_by_bin_uuid(model, bin_uuid):
    """Get an instance by binary uuid.

    :param model: a string, model name in rio.models.
    :param bin_uuid: a 16-bytes binary string.
    :return: None or a SQLAlchemy instance.
    """
    try:
        model = get_model(model)
    except ImportError:
        return None

    return model.query.filter_by(**{'bin_uuid': bin_uuid}).first()


def get_data_by_hex_uuid_or_404(model, hex_uuid, kind=''):
    """Get instance data by uuid and kind. Raise 404 Not Found if there is no data.

    This requires model has a `bin_uuid` column.

    :param model: a string, model name in rio.models
    :param hex_uuid: a hex uuid string in 24-bytes human-readable representation.
    :return: a dict.
    """
    uuid = UUID(hex_uuid)
    bin_uuid = uuid.get_bytes()

    instance = get_instance_by_bin_uuid(model, bin_uuid)

    if not instance:
        return abort(404)

    return ins2dict(instance, kind)



# -*- coding: utf-8 -*-
"""
rio.models
~~~~~~~~~~~~
"""

from .utils import get_data_or_404
from .utils import get_data_by_slug_or_404
from .utils import get_data_by_hex_uuid_or_404

from .user import User
from .project import Project
from .sender import Sender
from .action import Action
from .webhook import Webhook



"""init

Revision ID: 4cf90c101d82
Revises: None
Create Date: 2016-04-12 14:31:03.123085

"""

# revision identifiers, used by Alembic.
revision = '4cf90c101d82'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('user',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('confirmed_at', sa.DateTime(), nullable=True),
    sa.Column('username', sa.String(length=64), nullable=False),
    sa.Column('nickname', sa.String(length=32), server_default='', nullable=False),
    sa.Column('mobile', sa.String(length=11), nullable=True),
    sa.Column('email', sa.String(length=128), nullable=False),
    sa.Column('password', sa.String(length=128), server_default='', nullable=False),
    sa.Column('reset_password_token', sa.String(length=128), server_default='', nullable=False),
    sa.Column('auth_token', sa.String(length=128), nullable=True),
    sa.Column('active', sa.Boolean(), server_default='0', nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('auth_token', name='ux_user_auth_token'),
    sa.UniqueConstraint('email', name='ux_user_email'),
    sa.UniqueConstraint('mobile', name='ux_user_mobile'),
    sa.UniqueConstraint('username', name='ux_user_username')
    )
    op.create_table('project',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('slug', sa.String(length=64), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.Column('owner_id', sa.Integer(), nullable=False),
    sa.Column('status', sa.SmallInteger(), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('updated_at', sa.DateTime(), nullable=False),
    sa.ForeignKeyConstraint(['owner_id'], ['user.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('slug', name='ux_project_slug')
    )
    op.create_table('sender',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('slug', sa.String(length=64), nullable=False),
    sa.Column('token', sa.String(length=40), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('updated_at', sa.DateTime(), nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['project.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('project_id', 'slug', name='ux_sender_project_and_slug')
    )
    op.create_table('action',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('slug', sa.String(length=64), nullable=False),
    sa.Column('project_id', sa.Integer(), nullable=False),
    sa.Column('description', sa.String(length=255), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('updated_at', sa.DateTime(), nullable=False),
    sa.ForeignKeyConstraint(['project_id'], ['project.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('project_id', 'slug', name='ux_action_project_slug')
    )
    op.create_table('webhook',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('method_id', sa.SmallInteger(), nullable=False),
    sa.Column('action_id', sa.Integer(), nullable=False),
    sa.Column('raw_url', sa.String(length=1024), nullable=False),
    sa.Column('url_bin_digest', sa.BINARY(length=16), nullable=False),
    sa.Column('json_headers', sa.String(length=2048)),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('updated_at', sa.DateTime(), nullable=False),
    sa.ForeignKeyConstraint(['action_id'], ['action.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('action_id', 'url_bin_digest', name='ux_webhook_subscribe')
    )
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('webhook')
    op.drop_table('action')
    op.drop_table('sender')
    op.drop_table('project')
    op.drop_table('user')
    ### end Alembic commands ###



# -*- coding: utf-8 -*-
"""
rio._compact
~~~~~~~~~~~~~
"""

try:
    import cPickle as pickle
except ImportError:
    import pickle  # noqa

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO  # noqa

try:
    import simplejson as json
except ImportError:
    import json



# -*- coding: utf-8 -*-

DEBUG = True

SECRET_KEY = 'OOxdXBtiwPHGpjxaACWvzpYCbDhBmaYk'

SQLALCHEMY_DATABASE_URI = 'sqlite:////tmp/rio.db'

CELERY_BROKER_URL = 'redis://localhost/2'
CELERY_RESULT_BACKEND = 'db+' + SQLALCHEMY_DATABASE_URI

REDIS_DEFAULT_CLUSTERS = {
    0: {'host': '127.0.0.1', 'port': 6379},
}



# -*- coding: utf-8 -*-
"""
rio.models.project
~~~~~~~~~~~~~~~~~~
"""

from datetime import datetime

from rio.core import db

class Project(db.Model):
    """
    Projects are permission based namespaces which generally
    are the top level entry point for all data.
    """

    class Status:
        VISIBLE = 0
        HIDDEN = 1

    __table_name__ = 'project'
    __table_args__ = (
        db.UniqueConstraint('slug', name='ux_project_slug'),
    )

    id = db.Column(db.Integer(), primary_key=True)
    slug = db.Column(db.String(64), nullable=False)
    name = db.Column(db.String(128), nullable=False)
    owner_id = db.Column(db.Integer, db.ForeignKey('user.id'), nullable=False)
    status = db.Column(db.SmallInteger(), nullable=False, default=Status.VISIBLE)
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)

    actions = db.relationship('Action', backref=db.backref('project'), lazy='dynamic')
    senders = db.relationship('Sender', backref=db.backref('project'), lazy='dynamic')

    def __unicode__(self):
        return u'%s (%s)' % (self.name, self.slug)

    def to_simple_dict(self):
        return dict(
            id=self.id,
            slug=self.slug,
            name=self.name,
        )



# -*- coding: utf-8 -*-
"""
rio.exts.flask_redis_cache
~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import logging
from time import time

from flask import current_app
from werkzeug.utils import import_string

from rio._compat import json
from .base import Extension

logger = logging.getLogger('rio.exts.flask_cache')

class MemoryBackend(object):

    TABLE = {}
    EXPIRES = {}

    def get(self, key):
        if key not in self.EXPIRES or self.EXPIRES[key] > time():
            return self.TABLE.get(key)
        else:
            del self.TABLE[key]
            del self.EXPIRES[key]
            return

    def set(self, key, value):
        self.TABLE[key] = value
        if key in self.EXPIRES:
            del self.EXPIRES[key]

    def delete(self, key):
        if key in self.TABLE:
            del self.TABLE[key]
        if key in self.EXPIRES:
            del self.TABLE[key]

    def setex(self, key, timeout, value):
        self.TABLE[key] = value
        self.EXPIRES[key] = time() + timeout


class RedisBackend(object):

    def __init__(self, **kwargs):
        assert hasattr(current_app, 'extensions')
        cluster = current_app.extensions.get('rediscluster')
        assert cluster, 'Should init cluster before init cache.'
        self._client = cluster.default.get_routing_client()

    def get(self, key):
        """Execute `GET key` in redis."""
        return self._client.get(key)

    def set(self, key, value):
        """Execute `SET key value timeout` in redis. """
        return self._client.set(key, value)

    def delete(self, key):
        """Execute `DEL key` in redis."""
        return self._client.delete(key)

    def setex(self, key, timeout, value):
        return self._client.setex(key, timeout, value)


class Cache(Extension):

    def init_extension(self, app):
        """Initialize cache instance."""
        app.config.setdefault('CACHE_VERSION', '0')
        app.config.setdefault('CACHE_PREFIX', 'r')
        app.config.setdefault('CACHE_BACKEND', 'rio.exts.flask_cache.MemoryBackend')
        app.config.setdefault('CACHE_BACKEND_OPTIONS', {})

    @property
    def version(self):
        return current_app.config.get('CACHE_VERSION')

    @property
    def prefix(self):
        return current_app.config.get('CACHE_PREFIX')

    def make_key(self, key, version=None):
        """RedisCache will set prefix+version as prefix for each key."""
        return '{}:{}:{}'.format(
            self.prefix,
            version or self.version,
            key,
        )

    @property
    def client(self):
        """This method should return a client that implements:

        * set
        * get
        * setex
        * delete
        """
        return self.context('cacheclient', self.get_client)

    def get_client(self):
        """Get cache client.
        """
        backend_class = import_string(current_app.config.get('CACHE_BACKEND'))
        backend = backend_class(**current_app.config.get('CACHE_BACKEND_OPTIONS'))
        return backend


    def run(self, fn, ns='', version=None, **kwargs):
        kwargs_key = ':'.join('%s:%s' % (
            k, str(kwargs[k]).replace(' ', '')) for k in sorted(kwargs.keys()))
        key = '%s:%s:%s' % (ns, fn.__name__, kwargs_key)

        data = self.get(key, version)
        if data is not None:
            logger.debug('HIT %s', self.make_key(key))
            return data
        else:
            logger.debug('MISS %s', self.make_key(key))
            data = fn(**kwargs)
            if data is not None:
                logger.debug('CACHE %s', self.make_key(key))
                self.set(key, data, 86400, version)
            else:
                logger.debug('NIL %s', self.make_key(key))
            return data

    def set(self, key, value, timeout=None, version=None):
        key = self.make_key(key, version=version)
        v = json.dumps(value)
        if timeout:
            self.client.setex(key, int(timeout), v)
        else:
            self.client.set(key, v)

    def delete(self, key, version=None):
        key = self.make_key(key, version=version)
        self.client.delete(key)

    def get(self, key, version=None):
        key = self.make_key(key, version=version)
        result = self.client.get(key)
        if result is not None:
            result = json.loads(result)
        return result



# -*- coding: utf-8 -*-

from flask_script import Manager
from flask_migrate import MigrateCommand
from rio.app import create_app as create_flask
from rio.core import celery
from rio.core import migrate

app = create_flask()

manager = Manager(app)
manager.add_command('db', MigrateCommand)

def main():
    manager.run()



# -*- coding: utf-8 -*-

from pytest import fixture

from rio.core import cache


def test_cache_function_call(app):
    def hello_world(k1, k2):
        return {'k1': 'v1', 'k2': 'v2'}
    assert not cache.get(':hello_world:k1:v1:k2:v2')
    assert cache.run(hello_world, k1='v1', k2='v2')
    assert cache.get(':hello_world:k1:v1:k2:v2') == {'k1': 'v1', 'k2': 'v2'}



# -*- coding: utf-8 -*-

from base64 import b64encode

from flask import jsonify
from pytest import fixture

@fixture
def owner(session):
    from rio.models.user import User
    user = User(
        username='owner',
        email='owner@example.org',
    )
    session.add(user)
    session.commit()
    return user

@fixture
def project(session, owner):
    from rio.models.project import Project
    project_ = Project(
        owner_id=owner.id,
        slug='example-project',
        name='Example Project',
    )
    session.add(project_)
    session.commit()
    return project_

@fixture
def sender(session, project):
    from rio.models.sender import Sender
    sender_ = Sender(
        project_id=project.id,
        slug='example-sender',
        token='*',
    )
    session.add(sender_)
    session.commit()
    return sender_

@fixture
def sender_basic_token(sender):
    return b64encode('%s:%s' % (sender.slug, sender.token))

@fixture
def action(session, project):
    from rio.models.action import Action
    action_ = Action(
        project_id=project.id,
        slug='example-action',
    )
    session.add(action_)
    session.commit()
    return action_

@fixture
def builtin_webhook(app):
    @app.route('/webhook/success-example')
    def get():
        return jsonify(status='success', retval=0)

@fixture
def webhook(session, action):
    from rio.models.webhook import Webhook
    webhook_ = Webhook(
        action_id=action.id,
        method='GET',
        url='http://example.org'
    )
    session.add(webhook_)
    session.commit()
    return webhook_



# -*- coding: utf-8 -*-

import json

import pytest
import requests_mock
from flask import url_for


def test_emit_event_to_undefined_project_slug(client, sender_basic_token, action):
    url = url_for('event.emit_event', project_slug='non-exist', action_slug=action.slug)
    headers = {'Authorization': 'Basic %s' % sender_basic_token}
    resp = client.get(url, headers=headers)
    assert resp.status_code == 404

def test_emit_event_to_undefined_action_slug(client, sender_basic_token, project):
    url = url_for('event.emit_event', project_slug=project.slug, action_slug='non-exist')
    headers = {'Authorization': 'Basic %s' % sender_basic_token}
    resp = client.get(url, headers=headers)
    assert resp.status_code == 404


def test_emit_event_but_sender_has_not_logined(client, project, action):
    url = url_for('event.emit_event', project_slug=project.slug, action_slug=action.slug)
    resp = client.get(url)
    assert resp.status_code == 401


def test_emit_event_success(client, project, action, webhook, sender_basic_token):
    url = url_for('event.emit_event', project_slug=project.slug, action_slug=action.slug)
    headers = {'Authorization': 'Basic %s' % sender_basic_token}
    with requests_mock.Mocker() as m:
        m.get('http://example.org?key=value', text='data')
        resp = client.get(url+'?key=value', headers=headers)
    assert resp.status_code == 200
    data = json.loads(resp.data)
    assert data['message'] == 'ok'
    assert data['event']['uuid']


def test_emit_event_but_webhook_ran_failed(client, project, action, webhook, sender_basic_token):
    url = url_for('event.emit_event', project_slug=project.slug, action_slug=action.slug)
    headers = {'Authorization': 'Basic %s' % sender_basic_token}
    with requests_mock.Mocker() as m:
        m.get('http://example.org?key=value', text='failed', status_code=500)
        resp = client.get(url+'?key=value', headers=headers)
    assert resp.status_code == 200
    data = json.loads(resp.data)
    assert data['message'] == 'ok'
    assert data['event']['uuid']



# -*- coding: utf-8 -*-
"""
rio.blueprints.event.utils
~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from functools import wraps

from flask import request
from flask import jsonify

from rio.models  import validate_sender


def require_sender(f):
    """A decorator that protect emit view function is triggered by a trusted sender.

    Currently, Rio only support Basic Authorization.
    """
    @wraps(f)
    def decorator(*args, **kwargs):
        if not request.authorization:
            return jsonify({'message': 'unauthorized'}), 401

        username = request.authorization.username
        password = request.authorization.password

        if not validate_sender(username, password):
            return jsonify({'message': 'forbidden'}), 403

        return f(*args, **kwargs)
    return decorator



# -*- coding: utf-8 -*-

from datetime import datetime

from flask_user import UserMixin

from rio.core import db

class User(db.Model, UserMixin):

    __table_name__ = 'user'
    __table_args__ = (
        db.UniqueConstraint('username', name='ux_user_username'),
        db.UniqueConstraint('email', name='ux_user_email'),
        db.UniqueConstraint('mobile', name='ux_user_mobile'),
        db.UniqueConstraint('auth_token', name='ux_user_auth_token'),
    )

    id = db.Column(db.Integer(), primary_key=True)
    created_at = db.Column(db.DateTime(), default=datetime.utcnow)
    confirmed_at = db.Column(db.DateTime())
    username = db.Column(db.String(64), nullable=False)
    nickname = db.Column(db.String(32), nullable=False, server_default='')
    mobile = db.Column(db.String(11))
    email = db.Column(db.String(128), nullable=False)
    password = db.Column(db.String(128), nullable=False, server_default='')
    reset_password_token = db.Column(db.String(128), nullable=False, server_default='')
    auth_token = db.Column(db.String(128))
    active = db.Column(db.Boolean(), nullable=False, server_default='0')

    projects = db.relationship('Project', backref=db.backref('owner'), lazy='dynamic')

    def __unicode__(self):
        return 'user (%s)' % self.username



# -*- coding: utf-8 -*-
"""
rio.exts.flask_redis_cluster
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
from flask import current_app

from rb import Cluster

from .base import Extension



class RedisCluster(Extension):
    """Redis cluster extension."""

    def init_extension(self, app):
        app.config.setdefault('REDIS_DEFAULT_CLUSTERS', {
            0: {'port': 6379, 'host': '127.0.0.1'},
        })

    @property
    def default(self):
        return self.context('redis_defalut_cluster', self.make_redis_default_cluster)

    def make_redis_default_cluster(self):
        return Cluster(
            hosts=current_app.config.get('REDIS_DEFAULT_CLUSTERS'),
        )



# -*- coding: utf-8 -*-
"""
rio.setup
~~~~~~~~~
"""

from os import environ
from os import path

from .core import db
from .core import celery
from .core import redis
from .core import cache
from .core import sentry
from .core import migrate

def configure_app(app):
    """Configure Flask/Celery application.

    * Rio will find environment variable `RIO_SETTINGS` first::

        $ export RIO_SETTINGS=/path/to/settings.cfg
        $ rio worker

    * If `RIO_SETTINGS` is missing, Rio will try to load configuration
      module in `rio.settings` according to another environment
      variable `RIO_ENV`. Default load `rio.settings.dev`.

        $ export RIO_ENV=prod
        $ rio worker
    """
    app.config_from_object('rio.settings.default')

    if environ.get('RIO_SETTINGS'):
        app.config_from_envvar('RIO_SETTINGS')
        return

    config_map = {
        'dev': 'rio.settings.dev',
        'stag': 'rio.settings.stag',
        'prod': 'rio.settings.prod',
        'test': 'rio.settings.test',
    }

    rio_env = environ.get('RIO_ENV', 'dev')
    config = config_map.get(rio_env, config_map['dev'])
    app.config_from_object(config)


def register_blueprints(app):
    """Register blueprints to application.

    Currently, Rio registered:

    * /api/1
    * /dashboard
    """

    from .blueprints.event import bp as event_bp
    app.register_blueprint(event_bp, url_prefix='/event')

    from .blueprints.api_1 import bp as api_1_bp
    app.register_blueprint(api_1_bp, url_prefix='/api/1')

    from .blueprints.dashboard import bp as dashboard_bp
    app.register_blueprint(dashboard_bp, url_prefix='/dashboard')


def get_migration_assets():
    """Get absolute path of migration assets."""
    return path.join(path.dirname(__file__), 'migrations')


def init_core(app):
    """Init core objects."""
    from rio import models # noqa
    db.init_app(app)
    celery.init_app(app)
    redis.init_app(app)
    cache.init_app(app)
    sentry.init_app(app)
    migrate.init_app(app, db, directory=get_migration_assets())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'flask==0.10.1',
    'flask-sqlalchemy==2.1',
    'flask-migrate==1.8.0',
    'Flask-Celery-Helper==1.1.0',
    'flask-user==0.6.8',
    'celery==3.1.23',
    'redis==2.10.5',
    'rb==1.4',
    'simplejson==3.8.2',
    'blinker==1.4',
    'requests==2.9.1',
    'raven==5.12.0',
    'MySQL-python==1.2.5', # fixme
]

test_requirements = [
    'pytest==2.9.1',
    'requests-mock==0.7.0',
]

setup(
    name='rio',
    version='0.2.4',
    description="RESTful event dispatcher based on celery.",
    long_description=readme + '\n\n' + history,
    author="Ju Lin",
    author_email='soasme@gmail.com',
    url='https://github.com/soasme/rio',
    packages=find_packages(exclude=('tests', 'tests.*', '*.tests', '*.tests.*', )),
    package_dir={'rio': 'rio'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='rio',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    entry_points='''
        [console_scripts]
        rio=rio.manage:main
    '''
)



# -*- coding: utf-8 -*-
"""
rio.blueprints.event.views
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Implement of rio event view functions.
"""

from uuid import uuid4

from flask import jsonify
from flask import request

from rio.core import cache
from rio.tasks import exec_event
from rio.models  import get_data_by_slug_or_404
from .core import bp


@bp.errorhandler(404)
def not_found(error):
    """Error handler for 404."""
    return jsonify({'message': 'not found'}), 404


@bp.route('/<project_slug>/emit/<action_slug>', methods=['GET', 'POST'])
def emit_event(project_slug, action_slug):
    """Publish message to action.

    Rio will trigger all registered webhooks related to this action and
    trace running process.
    """
    # fetch project
    project = cache.run(get_data_by_slug_or_404,
                        model='project',
                        slug=project_slug,
                        kind='simple')
    project_id = project['id']

    # assert authorization
    if not request.authorization:
        return jsonify({'message': 'unauthorized'}), 401

    username = request.authorization.username
    password = request.authorization.password

    # assert sender
    sender = cache.run(get_data_by_slug_or_404,
                       model='sender',
                       slug=username,
                       kind='sensitive',
                       project_id=project_id,)

    if not sender:
        return jsonify({'message': 'no such sender'}), 401

    # assert sender token
    if sender['token'] != password:
        return jsonify({'message': 'wrong token'}), 401

    action = cache.run(get_data_by_slug_or_404,
                      model='action',
                      slug=action_slug,
                      kind='full',
                      project_id=project_id)

    # assert action belongs to a project
    if action['project']['slug'] != project['slug']:
        return jsonify(message='forbidden'), 403

    # execute event
    event = {'uuid': uuid4(), 'project': project['slug'], 'action': action['slug']}
    payload = request.values.to_dict()
    res = exec_event(event, action['webhooks'], payload)

    # response
    return jsonify(message='ok', task={'id': res.id}, event={'uuid': event['uuid']})



# -*- coding: utf-8 -*-
"""
rio.exts.base
~~~~~~~~~~~~~

Base implementation of flask extension.
"""

# Find the stack on which we want to store the database connection.
# Starting with Flask 0.9, the _app_ctx_stack is the correct one,
# before that we need to use the _request_ctx_stack.
try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack

class Extension(object):
    """Base extension."""

    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_extension(self, app):
        """Any inherited extension should implement this method.

        Normally do some configuration initialization stuff here.

        Example::

            def init_extension(self, app):
                app.config.setdefault('RIO_XXX', True)

        :param app: Flask application.
        """
        raise NotImplementedError

    def init_app(self, app):
        """Initialize extension to the given application.

        Extension will be registered to `app.extensions` with lower classname
        as key and instance as value.

        :param app: Flask application.
        """
        self.init_extension(app)

        if not hasattr(app, 'extensions'):
            app.extensions = {}
        classname = self.__class__.__name__
        extname = classname.replace('Flask', '').lower()
        app.extensions[extname] = self

    def context(self, key, method):
        """A helper method to attach a value within context.

        :param key: the key attached to the context.
        :param method: the constructor function.
        :return: the value attached to the context.
        """
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, key):
                setattr(ctx, key, method())
            return getattr(ctx, key)



# -*- coding: utf-8 -*-
"""
rio.core
~~~~~~~~~

Definition of rio core object.
"""

import logging

from flask_sqlalchemy import SQLAlchemy
from flask_celery import Celery
from flask_migrate import Migrate
from raven.contrib.flask import Sentry
from rio.exts.flask_redis_cluster import RedisCluster
from rio.exts.flask_cache import Cache

#: db: SQLAlchemy instance for database manipulation.
db = SQLAlchemy()

#: celery: Celery instance for job queue manipulation.
celery = Celery()

#: redis
redis = RedisCluster()

#: redis cache
cache = Cache()

#: logger instance
logger = logging.getLogger('rio')

#: sentry: Sentry Official Client
sentry = Sentry()

#: migrate: Database migration object
migrate = Migrate()



# -*- coding: utf-8 -*-

from datetime import datetime
import hashlib

from sqlalchemy.types import BINARY

from rio.core import db
from rio._compat import json
from .utils import ins2dict

class Webhook(db.Model):
    """Webhook Model.

    Webhook model defines the http method and url that will be triggered
    on receiving event.
    """

    __tablename__ = 'webhook'
    __table_args__ = (
        db.UniqueConstraint('action_id', 'url_bin_digest', name='ux_webhook_subscribe'),
    )

    class Method:
        """HTTP Methods.

        Currently, only GET/POST are supported.
        For database effencity, this field will be stored as integer.
        """
        GET = 1
        POST = 2
        MAP = {
            GET: 'GET',
            POST: 'POST',
        }

    id = db.Column(db.Integer(), primary_key=True)
    method_id = db.Column(db.SmallInteger(), nullable=False, default=Method.GET)
    action_id = db.Column(db.Integer(), db.ForeignKey('action.id'), nullable=False)
    url_bin_digest = db.Column(BINARY(16), nullable=False)
    raw_url = db.Column(db.String(1024), nullable=False)
    json_headers = db.Column(db.String(2048))
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)

    def __init__(self, action_id, url, method='GET', headers=None):
        self.action_id = action_id
        self.url = url
        self.method = method
        self.headers = headers or {}

    @staticmethod
    def generate_url_hash(url):
        return hashlib.md5(url).digest()

    @property
    def url(self):
        return self.raw_url

    @url.setter
    def url(self, value):
        self.raw_url = value
        self.url_bin_digest = self.generate_url_hash(value)

    @property
    def method(self):
        return self.Method.MAP[self.method_id]

    @method.setter
    def method(self, method):
        self.method_id = getattr(self.Method, method.upper())

    @property
    def headers(self):
        if not self.json_headers:
            return {}
        return json.loads(self.json_headers)

    @headers.setter
    def headers(self, headers):
        self.json_headers = json.dumps(headers)

    def to_full_dict(self):
        data = ins2dict(self)
        data.pop('action_id')
        data['action'] = ins2dict(getattr(self, 'action'), 'simple')
        return data

    def to_simple_dict(self):
        return dict(
            id=self.id,
            method=self.method,
            url=self.url,
            headers=self.headers or {},
        )



# -*- coding: utf-8 -*-

import logging

from requests import ConnectionError

from urlparse import urlparse
from rio.utils.http import urlopen

logger = logging.getLogger('rio.client')

def parse_netloc(netloc):
    auth, _netloc = netloc.split('@')
    sender, token = auth.split(':')
    domain, port = _netloc.split(':')
    port = int(port or 80)
    return dict(sender=sender, token=token, domain=domain, port=port)

def parse_path(path):
    version, project = path[1:].split('/')
    return dict(version=int(version), project=project)

def parse_dsn(dsn):
    parsed_dsn = urlparse(dsn)
    parsed_netloc = parse_netloc(parsed_dsn.netloc)
    parsed_path = parse_path(parsed_dsn.path)
    return {
        'scheme': parsed_dsn.scheme,
        'sender': parsed_netloc.get('sender'),
        'token': parsed_netloc.get('token'),
        'netloc': parsed_netloc.get('netloc'),
        'version': parsed_path.get('version'),
        'project': parsed_path.get('project'),
    }

class Client(object):

    def __init__(self, dsn):
        self.dsn = dsn
        self.context = parse_dsn(dsn)

    def emit(self, slug, payload):
        url = '%(scheme)s://%(domain)s:%(port)d/event/%(project)/emit/' % self.context
        url += slug
        return urlopen(url, 'POST', data=payload, timeout=3)



# -*- coding: utf-8 -*-
"""
rio.models.sender
~~~~~~~~~~~~~~~~~
"""

from datetime import datetime

from sqlalchemy import func

from rio.core import db

class Sender(db.Model):
    """to validate trusted senders."""

    __table_name__ = 'sender'
    __table_args__ = (
        db.UniqueConstraint('project_id', 'slug', name='ux_sender_project_and_slug'),
    )

    id = db.Column(db.Integer(), primary_key=True)
    project_id = db.Column(db.Integer(), db.ForeignKey('project.id'), nullable=False)
    slug = db.Column(db.String(64), nullable=False)
    token = db.Column(db.String(40), nullable=False)
    created_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)
    updated_at = db.Column(db.DateTime(), nullable=False, default=datetime.utcnow)

    def to_sensitive_dict(self):
        return dict(
            id=self.id,
            project_id=self.project_id,
            slug=self.slug,
            token=self.token,
        )
